from pathlib import Path

import torch
from ffmpy import FFExecutableNotFoundError, FFmpeg


def get_default_device():
  cuda_count = torch.cuda.device_count()
  if cuda_count == 1:
    return torch.device("cuda")
  if cuda_count > 1:
    return torch.device("cuda:0")
  return torch.device("cpu")


def get_sample_count(sampling_rate: int, duration_s: float):
  return int(round(sampling_rate * duration_s, 0))


def normalize_audio(path: Path, output: Path):
  ffmpeg_normalization = FFmpeg(
    inputs={
      str(path.absolute()): None
    },
    outputs={
      str(output.absolute()): "-acodec pcm_s16le -ar 22050 -ac 1 -af loudnorm=I=-16:LRA=11:TP=-1.5 -y -hide_banner -loglevel error"
    },
  )

  try:
    ffmpeg_normalization.run()
  except FFExecutableNotFoundError as error:
    raise Exception("FFmpeg was not found, therefore no normalization was applied!") from error
  except Exception as error:
    raise Exception(
      "FFmpeg couldn't be executed, therefore no normalization was applied!") from error



import logging
import os
from logging import Formatter, Handler, Logger, StreamHandler, getLogger
from logging.handlers import MemoryHandler
from pathlib import Path


class ConsoleFormatter(logging.Formatter):
  """Logging colored formatter, adapted from https://stackoverflow.com/a/56944256/3638629"""

  purple = '\x1b[34m'
  blue = '\x1b[36m'
  # blue = '\x1b[38;5;39m'
  yellow = '\x1b[38;5;226m'
  red = '\x1b[1;49;31m'
  bold_red = '\x1b[1;49;31m'
  reset = '\x1b[0m'

  def __init__(self):
    super().__init__()
    self.datefmt = '%H:%M:%S'
    fmt = '(%(levelname)s) %(message)s'
    fmt_info = '%(message)s'

    self.fmts = {
        logging.NOTSET: self.purple + fmt + self.reset,
        logging.DEBUG: self.blue + fmt + self.reset,
        logging.INFO: fmt_info,
        logging.WARNING: self.yellow + fmt + self.reset,
        logging.ERROR: self.red + fmt + self.reset,
        logging.CRITICAL: self.bold_red + fmt + self.reset,
    }

  def format(self, record):
    log_fmt = self.fmts.get(record.levelno)
    formatter = logging.Formatter(log_fmt, self.datefmt)

    return formatter.format(record)


def add_console_out(logger: Logger) -> StreamHandler:
  console = StreamHandler()
  logger.addHandler(console)
  set_console_formatter(console)
  return console


def set_console_formatter(handler: Handler) -> None:
  logging_formatter = ConsoleFormatter()
  handler.setFormatter(logging_formatter)


def set_logfile_formatter(handler: Handler) -> None:
  fmt = '[%(asctime)s.%(msecs)03d] %(name)s (%(levelname)s) %(message)s'
  datefmt = '%Y/%m/%d %H:%M:%S'
  logging_formatter = Formatter(fmt, datefmt)
  handler.setFormatter(logging_formatter)


def get_cli_logger() -> Logger:
  logger = getLogger("en_tts_cli")
  return logger


def get_file_logger() -> Logger:
  flogger = getLogger("file.en_tts_cli")
  if flogger.propagate:
    flogger.propagate = False
  return flogger


def configure_cli_logger(debug: bool = False) -> None:
  cli_logger = getLogger("en_tts_cli")
  cli_logger.handlers.clear()
  assert len(cli_logger.handlers) == 0
  console_handler = add_console_out(cli_logger)
  # console_handler.setLevel(logging.DEBUG if debug else logging.INFO)
  cli_logger.setLevel(logging.DEBUG if debug else logging.INFO)

  core_logger = getLogger("en_tts")
  core_logger.parent = cli_logger

  file_logger = get_file_logger()
  cli_logger.parent = file_logger


def configure_root_logger(debug: bool = False) -> None:
  # productive = False
  # loglevel = logging.INFO if productive else logging.DEBUG
  root_logger = getLogger()
  root_logger.setLevel(logging.DEBUG if debug else logging.INFO)
  root_logger.manager.disable = logging.NOTSET
  if len(root_logger.handlers) > 0:
    console = root_logger.handlers[0]
    set_console_formatter(console)
  else:
    console = add_console_out(root_logger)

  # console.setLevel(logging.DEBUG if debug else logging.INFO)


def configure_file_logger(path: Path, debug: bool = False, buffer_capacity: int = 1):
  flogger = get_file_logger()
  assert len(flogger.handlers) == 0
  path.parent.mkdir(parents=True, exist_ok=True)
  if path.is_file():
    os.remove(path)
  path.write_text("")
  fh = logging.FileHandler(path)
  set_logfile_formatter(fh)
  # fh.setLevel(logging.DEBUG if debug else logging.INFO)
  flogger.setLevel(logging.DEBUG if debug else logging.INFO)
  mh = MemoryHandler(buffer_capacity, logging.ERROR, fh, True)
  flogger.addHandler(mh)



from logging import getLogger

import numpy as np

from en_tts.helper import normalize_audio
from en_tts.io import save_audio
from en_tts.synthesizer import Synthesizer
from en_tts_tests.helper import get_tests_dir


def test_component():
  conf_dir = get_tests_dir()

  s = Synthesizer(conf_dir)

  text = 'ð|ˈɪ|s|SIL0|ˈɪ|z|SIL0|ə|SIL0|tː|ˈɛ|s|t|SIL0|ˈæ|b|?|SIL2\nə|n˘|d|SIL0|ˈaɪ˘|m|SIL0|ð|ˈɛr˘|SIL0|θ|ˈʌr|d˘|ˌi|-|wː|ˈʌː|nː|.|SIL2'

  audio = s.synthesize(text)
  save_audio(audio, conf_dir / "output.wav")
  normalize_audio(conf_dir / "output.wav", conf_dir / "output_norm.wav")
  logger = getLogger(__name__)
  logger.info(conf_dir / "output_norm.wav")
  assert len(audio) > 0

  np.testing.assert_array_almost_equal(
    audio[:10],
    np.array(
      [
        4.06468927e-04, -1.38543465e-03, -2.93170044e-04, -3.17878075e-05,
        -1.35282415e-03, -1.24711674e-04, -6.18043647e-04, -2.78891705e-04,
        9.47587105e-05, -4.59646282e-04
      ],
      dtype=np.float64,
    )
  )
  assert audio.dtype == np.float64
  assert audio.shape == (58682,)


def test_empty():
  conf_dir = get_tests_dir()

  s = Synthesizer(conf_dir)

  text = ''

  audio = s.synthesize(text)
  assert audio.dtype == np.float64
  assert audio.shape == (0,)



from pathlib import Path
from tempfile import gettempdir


def get_conf_dir() -> Path:
  conf_dir = Path.home() / ".en-tts"
  return conf_dir


def get_work_dir() -> Path:
  work_dir = Path(gettempdir()) / "en-tts"
  return work_dir



import logging
import re
from collections import OrderedDict
from copy import deepcopy
from logging import getLogger
from pathlib import Path
from typing import Generator, Optional

from dict_from_dict import create_dict_from_dict
from english_text_normalization import (add_space_around_dashes, execute_pipeline,
                                        expand_abbreviations, expand_and_a_half, geo_to_george,
                                        normalize_all_units, normalize_am_and_pm,
                                        normalize_degrees_minutes_and_seconds,
                                        normalize_emails_and_at,
                                        normalize_king_name_followed_by_roman_numeral,
                                        normalize_numbers, normalize_percent,
                                        normalize_pounds_shillings_and_pence,
                                        normalize_second_and_third_when_abbr_with_d,
                                        normalize_temperatures_general,
                                        normalize_three_and_four_dots,
                                        normalize_today_tomorrow_and_tonight,
                                        remove_dot_after_single_capital_letters,
                                        remove_double_hyphen_before_or_after_colon,
                                        remove_equal_sign, remove_sic, remove_stars,
                                        remove_underscore_characters,
                                        remove_whitespace_before_sentence_punctuation,
                                        replace_and_sign_with_word_and, replace_eg_with_for_example,
                                        replace_etc_with_et_cetera, replace_four_hyphens_by_two,
                                        replace_hyphen_between_numbers_with_to,
                                        replace_ie_with_that_is, replace_no_with_number,
                                        replace_nos_with_numbers, strip,
                                        write_out_month_abbreviations)
from ordered_set import OrderedSet
from pronunciation_dictionary import MultiprocessingOptions, PronunciationDict
from pronunciation_dictionary_utils import (map_symbols_dict, merge_dictionaries,
                                            replace_symbols_in_pronunciations,
                                            select_single_pronunciation)
from txt_utils import extract_vocabulary_from_text, replace_text, transcribe_text_using_dict
from unidecode import unidecode_expect_ascii

from en_tts.globals import DEFAULT_CONF_DIR
from en_tts.resources import download_nltk_data, get_cmu_dict, get_ljs_dict

ARPA_IPA_MAPPING = {
  "AO0": "ɔ",
  "AO1": "ˈɔ",
  "AO2": "ˌɔ",
  "EY0": "eɪ",
  "EY1": "ˈeɪ",
  "EY2": "ˌeɪ",
  "UW0": "u",
  "UW1": "ˈu",
  "UW2": "ˌu",
  "ER0": "ər",  # !,
  "ER1": "ˈʌr",  # !,
  "ER2": "ˌʌr",  # !,
  "IH0": "ɪ",
  "IH1": "ˈɪ",
  "IH2": "ˌɪ",
  "EH0": "ɛ",
  "EH1": "ˈɛ",
  "EH2": "ˌɛ",
  "IY0": "i",
  "IY1": "ˈi",
  "IY2": "ˌi",
  "AA0": "ɑ",
  "AA1": "ˈɑ",
  "AA2": "ˌɑ",
  "AE0": "æ",
  "AE1": "ˈæ",
  "AE2": "ˌæ",
  "OW0": "oʊ",
  "OW1": "ˈoʊ",
  "OW2": "ˌoʊ",
  "AY0": "aɪ",
  "AY1": "ˈaɪ",
  "AY2": "ˌaɪ",
  "AW0": "aʊ",
  "AW1": "ˈaʊ",
  "AW2": "ˌaʊ",
  "AH0": "ə",  # !,
  "AH1": "ˈʌ",
  "AH2": "ˌʌ",
  "OY0": "ɔɪ",
  "OY1": "ˈɔɪ",
  "OY2": "ˌɔɪ",
  "UH0": "ʊ",
  "UH1": "ˈʊ",
  "UH2": "ˌʊ",
  "SH": "ʃ",
  "B": "b",
  "D": "d",
  "M": "m",
  "Q": "ʔ",
  "L": "l",
  "K": "k",
  "R": "r",  # ɹ,
  "CH": "tʃ",
  "NG": "ŋ",
  "P": "p",
  "Z": "z",
  "ZH": "ʒ",
  "JH": "dʒ",
  "T": "t",
  "G": "ɡ",
  "N": "n",
  "F": "f",
  "TH": "θ",
  "V": "v",
  "HH": "h",
  "DH": "ð",
  "S": "s",
  "W": "w",
  "Y": "j",
}


class Transcriber():
  def __init__(
      self,
      conf_dir: Path = DEFAULT_CONF_DIR,
  ) -> None:
    logger = getLogger(__name__)
    tmp_logger = getLogger("english_text_normalization")
    tmp_logger.parent = logger
    tmp_logger.setLevel(logging.WARNING)

    tmp_logger = getLogger("dict_from_dict")
    tmp_logger.parent = logger
    tmp_logger.setLevel(logging.WARNING)

    tmp_logger = getLogger("dict_from_g2pE")
    tmp_logger.parent = logger
    tmp_logger.setLevel(logging.WARNING)

    tmp_logger = getLogger("pronunciation_dictionary")
    tmp_logger.parent = logger
    tmp_logger.setLevel(logging.WARNING)

    tmp_logger = getLogger("txt_utils")
    tmp_logger.parent = logger
    tmp_logger.setLevel(logging.WARNING)

    tmp_logger = getLogger("pronunciation_dictionary_utils")
    tmp_logger.parent = logger
    tmp_logger.setLevel(logging.WARNING)

    download_nltk_data()

    self._conf_dir = conf_dir
    self._ljs_dict = get_ljs_dict(conf_dir)
    self._cmu_dict = get_cmu_dict(conf_dir)
    self._symbol_separator = "|"
    self._punctuation = {".", "!", "?", ",", ":", ";", "\"", "'", "[", "]", "(", ")", "-", "—"}
    self.text_normed: Optional[str] = None
    self.text_sentenced: Optional[str] = None
    self.vocabulary: OrderedSet[str] = OrderedSet()
    self.dict1: PronunciationDict = OrderedDict()
    self.oov1: Optional[OrderedSet[str]] = None
    self.dict1_single: Optional[PronunciationDict] = None
    self.dict2: Optional[PronunciationDict] = None
    self.dict2_single: Optional[PronunciationDict] = None
    self.oov2: Optional[OrderedSet[str]] = None
    self.dict1_2: Optional[PronunciationDict] = None
    self.dict3: Optional[PronunciationDict] = None
    self.dict3_single: Optional[PronunciationDict] = None
    self.oov3: Optional[OrderedSet[str]] = None
    self.dict1_2_3: Optional[PronunciationDict] = None
    self.dict4_arpa: Optional[PronunciationDict] = None
    self.dict4: Optional[PronunciationDict] = None
    self.dict4_single: Optional[PronunciationDict] = None
    self.dict1_2_3_4: Optional[PronunciationDict] = None
    self.text_ipa: str = ""
    self.text_ipa_readable: str = ""

  def _reset_locals(self) -> None:
    self.text_normed = None
    self.text_sentenced = None
    self.vocabulary = OrderedSet()
    self.dict1 = OrderedSet()
    self.oov1 = None
    self.dict1_single = None
    self.dict2 = None
    self.dict2_single = None
    self.oov2 = None
    self.dict1_2 = None
    self.dict3 = None
    self.dict3_single = None
    self.oov3 = None
    self.dict1_2_3 = None
    self.dict4_arpa = None
    self.dict4 = None
    self.dict4_single = None
    self.dict1_2_3_4 = None
    self.text_ipa = ""
    self.text_ipa_readable = ""

  def transcribe_to_ipa(self, text: str, skip_normalization: bool = False, skip_sentence_separation: bool = False) -> str:
    logger = getLogger(__name__)
    self._reset_locals()

    if skip_normalization:
      logger.debug("Normalization was skipped.")
    else:
      logger.info("Normalizing ...")
      text_normed = normalize_eng_text(text)
      if text_normed == text:
        logger.debug("No normalization applied.")
      else:
        self.text_normed = text_normed
        text = text_normed
        logger.debug("Normalization was applied.")

    if skip_sentence_separation:
      logger.debug("Sentence separation was skipped.")
    else:
      logger.info("Separating sentences ...")
      sentences = get_sentences(text)
      text_sentenced = "\n".join(sentences)
      if text == text_sentenced:
        logger.debug("No sentence separation applied.")
      else:
        self.text_sentenced = text_sentenced
        text = text_sentenced
        logger.debug("Sentence separation was applied.")

    logger.debug("Extracting vocabulary ...")
    vocabulary = extract_vocabulary_from_text(text, n_jobs=1, chunksize=2_000_000, silent=True)
    self.vocabulary = vocabulary

    logger.info("Looking up vocabulary ...")
    dict1, oov1 = create_dict_from_dict(vocabulary, self._ljs_dict, trim={
    }, split_on_hyphen=False, ignore_case=False, n_jobs=1, maxtasksperchild=None, chunksize=10_000, silent=True)

    self.dict1 = deepcopy(dict1)
    if len(oov1) > 0:
      self.oov1 = oov1

    changed_word_count = select_single_pronunciation(dict1, mode="highest-weight", seed=None,
                                                     mp_options=MultiprocessingOptions(1, None, 1_000), silent=True)

    self.dict1_single = None
    if changed_word_count > 0:
      self.dict1_single = deepcopy(dict1)

    oov2: OrderedSet[str] = OrderedSet()
    if len(oov1) > 0:
      dict2, oov2 = create_dict_from_dict(oov1, self._ljs_dict, trim=self._punctuation, split_on_hyphen=True,
                                          ignore_case=True, n_jobs=1, maxtasksperchild=None, chunksize=10_000, silent=True)
      self.dict2 = deepcopy(dict2)
      if len(oov2) > 0:
        self.oov2 = oov2

      changed_word_count = select_single_pronunciation(dict2, mode="highest-weight", seed=None,
                                                       mp_options=MultiprocessingOptions(1, None, 1_000), silent=True)
      if changed_word_count > 0:
        self.dict2_single = deepcopy(dict2)

      merge_dictionaries(dict1, dict2, mode="add")

      self.dict1_2 = deepcopy(dict1)

    oov3: OrderedSet[str] = OrderedSet()
    if len(oov2) > 0:
      dict3, oov3 = create_dict_from_dict(oov2, self._cmu_dict, trim=self._punctuation, split_on_hyphen=True,
                                          ignore_case=True, n_jobs=1, maxtasksperchild=None, chunksize=10_000, silent=True)
      self.dict3 = deepcopy(dict3)
      if len(oov3) > 0:
        self.oov3 = oov3

      changed_word_count = select_single_pronunciation(dict3, mode="highest-weight", seed=None,
                                                       mp_options=MultiprocessingOptions(1, None, 1_000), silent=True)
      if changed_word_count > 0:
        self.dict3_single = deepcopy(dict3)

      merge_dictionaries(dict1, dict3, mode="add")
      self.dict1_2_3 = deepcopy(dict1)

    if len(oov3) > 0:
      from dict_from_g2pE import transcribe_with_g2pE
      dict4 = transcribe_with_g2pE(oov3, weight=1, trim=self._punctuation,
                                   split_on_hyphen=True, n_jobs=1, maxtasksperchild=None, chunksize=100_000, silent=True)
      self.dict4_arpa = deepcopy(dict4)

      map_symbols_dict(dict4, ARPA_IPA_MAPPING, partial_mapping=False,
                       mp_options=MultiprocessingOptions(1, None, 100_000), silent=True)
      replace_symbols_in_pronunciations(dict4, "( |ˈ|ˌ)(ə|ʌ|ɔ|ɪ|ɛ|ʊ) r", r"\1\2r",
                                        False, None, MultiprocessingOptions(1, None, 100_000), silent=True)
      self.dict4 = deepcopy(dict4)

      changed_word_count = select_single_pronunciation(dict4, mode="highest-weight", seed=None,
                                                       mp_options=MultiprocessingOptions(1, None, 1_000), silent=True)
      if changed_word_count > 0:
        self.dict4_single = deepcopy(dict4)

      merge_dictionaries(dict1, dict4, mode="add")

      self.dict1_2_3_4 = deepcopy(dict1)

    logger.debug("Transcribing to IPA ...")
    text_ipa = transcribe_text_using_dict(
      text, dict1,
      phoneme_sep=self._symbol_separator, n_jobs=1, chunksize=2_000_000, silent=True,
    )
    self.text_ipa = text_ipa
    self.text_ipa_readable = text_ipa.replace(self._symbol_separator, "")

    text_ipa = replace_text(text_ipa, " ", "SIL0", disable_regex=True)
    text_ipa = replace_text(text_ipa, f",{self._symbol_separator}SIL0",
                            f",{self._symbol_separator}SIL1", disable_regex=True)
    text_ipa = replace_text(text_ipa, r"(\.|\!|\?)",
                            rf"\1{self._symbol_separator}SIL2", disable_regex=False)
    text_ipa = replace_text(
      text_ipa, rf"(;|:){re.escape(self._symbol_separator)}SIL0", rf"\1{self._symbol_separator}SIL2", disable_regex=False)

    return text_ipa


def get_sentences(text: str) -> Generator[str, None, None]:
  pattern = re.compile(r"(\.|\?|\!) +")  # [^$]
  sentences = pattern.split(text)
  for i in range(0, len(sentences), 2):
    if i + 1 < len(sentences):
      sentence = sentences[i] + sentences[i + 1]
    else:
      sentence = sentences[i]
    if len(sentence) > 0:
      yield sentence


def normalize_eng_text(text: str) -> str:
  text = execute_pipeline(
    text,
    (
      normalize_emails_and_at,
      remove_underscore_characters,
      remove_equal_sign,
      add_space_around_dashes,
      replace_ie_with_that_is,
      replace_eg_with_for_example,
      replace_etc_with_et_cetera,
      replace_nos_with_numbers,
      replace_no_with_number,
      remove_urls,
      geo_to_george,
      write_out_month_abbreviations,
      normalize_today_tomorrow_and_tonight,
      normalize_king_name_followed_by_roman_numeral,
      normalize_am_and_pm,
      normalize_pounds_shillings_and_pence,
      normalize_temperatures_general,
      normalize_degrees_minutes_and_seconds,
      normalize_all_units,
      normalize_percent,
      expand_and_a_half,
      replace_hyphen_between_numbers_with_to,
      normalize_second_and_third_when_abbr_with_d,
      normalize_numbers,
      expand_abbreviations,
      remove_dot_after_single_capital_letters,
      replace_and_sign_with_word_and,
      remove_double_hyphen_before_or_after_colon,
      normalize_three_and_four_dots,
      replace_four_hyphens_by_two,
      add_space_around_dashes,
      remove_sic,
      remove_stars,
      remove_whitespace_before_sentence_punctuation,
      strip,
      unidecode_expect_ascii,
      change_minus,
    )
  )

  return text


def change_minus(text: str) -> str:
  # "- -" to "—"
  text = text.replace(" -- ", " — ")
  text = text.replace(" - ", " — ")
  return text


def remove_urls(text: str) -> str:
  pattern = re.compile(
    r"(http|ftp|https)://([\w_-]+(?:(?:\.[\w_-]+)+))([\w.,@?^=%&:/~+#-]*[\w@?^=%&/~+#-])")
  result = pattern.sub("U R L", text)
  return result



from en_tts.helper import get_default_device
from en_tts_cli.cli import reset_work_dir
from en_tts_cli.globals import get_work_dir
from en_tts_cli.main import synthesize_ipa


def test_component():
  text = 'ð|ˈɪ|s|SIL0|ˈɪ|z|SIL0|ə|SIL0|tː|ˈɛ|s|t|SIL0|ˈæ|b|?|SIL2\nə|n˘|d|SIL0|ˈaɪ˘|m|SIL0|ð|ˈɛr˘|SIL0|θ|ˈʌr|d˘|ˌi|-|wː|ˈʌː|nː|.|SIL2'
  work_dir = get_work_dir()

  reset_work_dir()
  synthesize_ipa(text, 5000, 1.0, 0.0005, 0, get_default_device(),
                 0.2, 1.0, 2, work_dir / "output.wav")



from en_tts.helper import get_default_device
from en_tts_cli.cli import reset_work_dir
from en_tts_cli.globals import get_work_dir
from en_tts_cli.main import synthesize_english


def test_component():
  text = "This is a test abbb? And I'm there 31."
  work_dir = get_work_dir()

  reset_work_dir()
  synthesize_english(text, 5000, 1.0, 0.0005, 0, get_default_device(),
                     0.2, 1.0, 2, False, False, work_dir / "output.wav")




import os
from pathlib import Path
from typing import Generator


def get_all_files_in_all_subfolders(directory: Path) -> Generator[Path, None, None]:
  for root, _, files in os.walk(directory):
    for name in files:
      file_path = Path(root) / name
      yield file_path



import pickle
from pathlib import Path
from typing import Any

import numpy as np
from waveglow import float_to_wav


def save_obj(obj: Any, path: Path) -> None:
  assert isinstance(path, Path)
  assert path.parent.exists() and path.parent.is_dir()
  with open(path, mode="wb") as file:
    pickle.dump(obj, file)


def load_obj(path: Path) -> Any:
  assert isinstance(path, Path)
  assert path.is_file()
  with open(path, mode="rb") as file:
    return pickle.load(file)


def save_audio(wav: np.ndarray, path: Path):
  float_to_wav(wav, str(path.absolute()), sample_rate=22050)



import logging
import shutil
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Callable, Optional

import torch
from ordered_set import OrderedSet
from pronunciation_dictionary import PronunciationDict, SerializationOptions, save_dict

from en_tts.helper import get_default_device, normalize_audio
from en_tts.io import save_audio
from en_tts.synthesizer import Synthesizer
from en_tts.transcriber import Transcriber
from en_tts_cli.argparse_helper import (get_torch_devices, parse_device,
                                        parse_float_between_zero_and_one,
                                        parse_non_empty_or_whitespace, parse_non_negative_float,
                                        parse_non_negative_integer, parse_path,
                                        parse_positive_integer)
from en_tts_cli.globals import get_conf_dir, get_work_dir
from en_tts_cli.logging_configuration import get_file_logger


def init_synthesize_eng_parser(parser: ArgumentParser) -> Callable[[Namespace], None]:
  parser.description = "Synthesize English texts into speech."
  parser.add_argument("input", type=parse_non_empty_or_whitespace, metavar="INPUT",
                      help="text input")
  parser.add_argument("--skip-normalization", action="store_true", help="skip normalization step")
  parser.add_argument("--skip-sentence-separation", action="store_true",
                      help="skip sentence separation step")
  add_common_arguments(parser)

  def parse_ns(ns: Namespace) -> None:
    synthesize_english(ns.input, ns.max_decoder_steps, ns.sigma, ns.denoiser_strength, ns.seed, ns.device,
                       ns.silence_sentences, ns.silence_paragraphs, ns.loglevel, ns.skip_normalization, ns.skip_sentence_separation, ns.output)
  return parse_ns


def init_synthesize_ipa_parser(parser: ArgumentParser) -> Callable[[Namespace], None]:
  parser.description = "Synthesize English IPA-transcribed texts into speech."
  parser.add_argument("input", type=parse_non_empty_or_whitespace, metavar="INPUT",
                      help="text input")
  add_common_arguments(parser)

  def parse_ns(ns: Namespace) -> None:
    synthesize_ipa(ns.input, ns.max_decoder_steps, ns.sigma, ns.denoiser_strength, ns.seed,
                   ns.device, ns.silence_sentences, ns.silence_paragraphs, ns.loglevel, ns.output)

  return parse_ns


def add_common_arguments(parser: ArgumentParser) -> None:
  parser.add_argument("--silence-sentences", metavar="SECONDS", type=parse_non_negative_float,
                      help="add silence between sentences (in seconds)", default=0.2)
  parser.add_argument("--silence-paragraphs", metavar="SECONDS", type=parse_non_negative_float,
                      help="add silence between paragraphs (in seconds)", default=1.0)
  parser.add_argument("--seed", type=parse_non_negative_integer, metavar="SEED",
                      help="seed for generating speech", default=0)
  add_device_argument(parser)
  add_max_decoder_steps_argument(parser)
  add_denoiser_and_sigma_arguments(parser)
  parser.add_argument("--output", type=parse_path, metavar="PATH",
                      help="save audio to this location", default=get_work_dir() / "output.wav")


def add_denoiser_and_sigma_arguments(parser: ArgumentParser) -> None:
  parser.add_argument("--sigma", metavar="SIGMA", type=parse_float_between_zero_and_one,
                      default=1.0, help="sigma used for WaveGlow synthesis")
  parser.add_argument("--denoiser-strength", metavar="STRENGTH", default=0.0005,
                      type=parse_float_between_zero_and_one, help='strength of denoising to remove model bias after WaveGlow synthesis')


def add_max_decoder_steps_argument(parser: ArgumentParser) -> None:
  parser.add_argument('--max-decoder-steps', type=parse_positive_integer, metavar="STEPS",
                      default=5000, help="maximum step count before synthesis is stopped")


def add_device_argument(parser: ArgumentParser) -> None:
  parser.add_argument("--device", choices=list(get_torch_devices()), type=parse_device,
                      default=get_default_device(), help="use this device")


def synthesize_english(text: str, max_decoder_steps: int, sigma: float, denoiser_strength: float, seed: int, device: torch.device, silence_sentences: float, silence_paragraphs: float, loglevel: int, skip_normalization: bool, skip_sentence_separation: bool, output: Path):
  if loglevel == 0:
    cli_logger = logging.getLogger("en_tts_cli")
    cli_logger.setLevel(logging.WARNING)

  text_ipa = convert_eng_to_ipa(text, loglevel, skip_normalization, skip_sentence_separation)
  synthesize_ipa_core(text_ipa, max_decoder_steps, sigma, denoiser_strength,
                      seed, device, silence_sentences, silence_paragraphs, loglevel, output)


def synthesize_ipa(text_ipa: str, max_decoder_steps: int, sigma: float, denoiser_strength: float, seed: int, device: torch.device, silence_sentences: float, silence_paragraphs: float, loglevel: int, output: Path):
  if loglevel == 0:
    cli_logger = logging.getLogger("en_tts_cli")
    cli_logger.setLevel(logging.WARNING)

  if loglevel >= 1:
    try_log_text(text_ipa, "text")

  synthesize_ipa_core(text_ipa, max_decoder_steps, sigma, denoiser_strength,
                      seed, device, silence_sentences, silence_paragraphs, loglevel, output)


def convert_eng_to_ipa(text: str, loglevel: int, skip_normalization: bool, skip_sentence_separation: bool) -> str:
  conf_dir = get_conf_dir()

  t = Transcriber(conf_dir)

  text_ipa = t.transcribe_to_ipa(text, skip_normalization, skip_sentence_separation)

  if loglevel >= 1:
    for txt, name in (
      (text, "text"),
      (t.text_normed, "text.normed"),
      (t.text_sentenced, "text.sentenced"),
      (t.text_ipa, "text.ipa"),
      (t.text_ipa_readable, "text.ipa.readable"),
      (text_ipa, "text.ipa.silenced"),
    ):
      try_log_text(txt, name)

    for v, name in (
      (t.vocabulary, "vocabulary"),
      (t.oov1, "oov1"),
      (t.oov2, "oov2"),
      (t.oov3, "oov3"),
    ):
      try_log_voc(v, name)

    for d, name in (
      (t.dict1, "dict1"),
      (t.dict1_single, "dict1.single"),
      (t.dict2, "dict2"),
      (t.dict2_single, "dict2.single"),
      (t.dict1_2, "dict1+2"),
      (t.dict3, "dict3"),
      (t.dict3_single, "dict3.single"),
      (t.dict1_2_3, "dict1+2+3"),
      (t.dict4_arpa, "dict4.arpa"),
      (t.dict4, "dict4"),
      (t.dict4_single, "dict4.single"),
      (t.dict1_2_3_4, "dict1+2+3+4"),
    ):
      try_log_dict(d, name)
  return text_ipa


def synthesize_ipa_core(text_ipa: str, max_decoder_steps: int, sigma: float, denoiser_strength: float, seed: int, device: torch.device, silence_sentences: float, silence_paragraphs: float, loglevel: int, output: Path):
  logger = logging.getLogger(__name__)
  conf_dir = get_conf_dir()
  work_dir = get_work_dir()

  synthesizer = Synthesizer(conf_dir, device)
  audio = synthesizer.synthesize(text_ipa, max_decoder_steps, seed, sigma,
                                 denoiser_strength, silence_sentences, silence_paragraphs, silent=loglevel == 0)
  unnormed_out = work_dir / "output.unnormed.wav"
  save_audio(audio, unnormed_out)
  work_dir_output = unnormed_out

  try:
    normalize_audio(unnormed_out, work_dir / "output.wav")
    work_dir_output = work_dir / "output.wav"
  except Exception as error:
    logger.warning("Normalization was not possible!", exc_info=error, stack_info=True)
    logger.info(f"Saved audio to: '{unnormed_out.absolute()}'")

  if output != work_dir_output:
    try:
      output.parent.mkdir(parents=True, exist_ok=True)
      shutil.copyfile(work_dir_output, output)
    except Exception as ex:
      logger.exception(
        f"Output couldn't be created at: '{output.absolute()}'", exc_info=ex, stack_info=True)
      logger.info(f"Saved audio to: '{work_dir_output.absolute()}'")
  logger.info(f"Saved audio to: '{output.absolute()}'")


def try_log_dict(dictionary: Optional[PronunciationDict], name: str) -> None:
  if dictionary:
    work_dir = get_work_dir()
    logfile = work_dir / f"{name}.dict"
    save_dict(dictionary, logfile, "utf-8", SerializationOptions("TAB", False, True))
    flogger = get_file_logger()
    flogger.info(f"{name}: {logfile.absolute()}")


def try_log_voc(vocabulary: Optional[OrderedSet[str]], name: str) -> None:
  if vocabulary:
    work_dir = get_work_dir()
    logfile = work_dir / f"{name}.txt"
    logfile.write_text("\n".join(vocabulary), "utf-8")
    flogger = get_file_logger()
    flogger.info(f"{name}: {logfile.absolute()}")


def try_log_text(text: Optional[str], name: str) -> None:
  if text:
    work_dir = get_work_dir()
    logfile = work_dir / f"{name}.txt"
    logfile.write_text(text, "utf-8")
    flogger = get_file_logger()
    flogger.info(f"{name}: {logfile.absolute()}")



import argparse
import logging
import platform
import shutil
import sys
from argparse import ArgumentParser
from importlib.metadata import version
from logging import getLogger
from pathlib import Path
from pkgutil import iter_modules
from tempfile import gettempdir
from time import perf_counter
from typing import Callable, Generator, List, Tuple, cast

from en_tts_cli.globals import get_conf_dir, get_work_dir
from en_tts_cli.logging_configuration import (configure_cli_logger, configure_file_logger,
                                              configure_root_logger, get_file_logger)
from en_tts_cli.main import init_synthesize_eng_parser, init_synthesize_ipa_parser

__APP_NAME = "en-tts"

__version__ = version(__APP_NAME)

INVOKE_HANDLER_VAR = "invoke_handler"


def formatter(prog):
  return argparse.ArgumentDefaultsHelpFormatter(prog, max_help_position=40)


def print_features():
  parsers = get_parsers()
  for command, description, method in parsers:
    print(f"- `{command}`: {description}")


def get_parsers() -> Generator[Tuple[str, str, Callable], None, None]:
  yield from (
    ("synthesize", "synthesize English texts", init_synthesize_eng_parser),
    ("synthesize-ipa", "synthesize English texts transcribed in IPA", init_synthesize_ipa_parser),
  )


def _init_parser():
  main_parser = ArgumentParser(
    formatter_class=formatter,
    description="Command-line interface for synthesizing English texts into speech.",
  )
  main_parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + __version__)
  subparsers = main_parser.add_subparsers(help="description")

  for command, description, method in get_parsers():
    method_parser = subparsers.add_parser(
      command, help=description, formatter_class=formatter)
    # init parser
    invoke_method = method(method_parser)
    method_parser.set_defaults(**{
      INVOKE_HANDLER_VAR: invoke_method,
    })

    logging_group = method_parser.add_argument_group("logging arguments")
    # logging_group.add_argument("--work-directory", type=parse_path, metavar="DIRECTORY",
    #                            help="path to write the log", default=Path(gettempdir()) / "en-tts")
    logging_group.add_argument("--loglevel", metavar="LEVEL", type=int,
                               choices=[0, 1, 2], help="log-level", default=1)
    logging_group.add_argument("--debug", action="store_true",
                               help="include debugging information in log")

  return main_parser


def reset_work_dir():
  root_logger = getLogger()
  work_dir = get_work_dir()

  if work_dir.is_dir():
    root_logger.debug("Deleting working directory ...")
    shutil.rmtree(work_dir)
  root_logger.debug("Creating working directory ...")
  work_dir.mkdir(parents=False, exist_ok=False)


def ensure_conf_dir_exists():
  conf_dir = get_conf_dir()
  if not conf_dir.is_dir():
    root_logger = getLogger()
    root_logger.debug("Creating configuration directory ...")
    conf_dir.mkdir(parents=False, exist_ok=False)


def parse_args(args: List[str]) -> None:
  local_debugging = debug_file_exists()

  configure_root_logger(local_debugging)
  root_logger = getLogger()
  root_logger.debug(f"Received arguments: {str(args)}")

  parser = _init_parser()

  try:
    ns = parser.parse_args(args)
  except SystemExit as error:
    error_code = error.args[0]
    # -v -> 0; invalid arg -> 2
    sys.exit(error_code)

  root_logger.debug(f"Parsed arguments: {str(ns)}")

  if not hasattr(ns, INVOKE_HANDLER_VAR):
    parser.print_help()
    sys.exit(0)

  debug = cast(bool, ns.debug)
  root_logger.setLevel(logging.DEBUG if debug else logging.INFO)

  invoke_handler: Callable[..., bool] = getattr(ns, INVOKE_HANDLER_VAR)
  delattr(ns, INVOKE_HANDLER_VAR)

  ensure_conf_dir_exists()

  try:
    reset_work_dir()
  except Exception as ex:
    root_logger.exception("Working directory couldn't be resetted!", exc_info=ex, stack_info=True)
    sys.exit(1)

  work_dir = get_work_dir()
  logfile = work_dir / "output.log"
  try:
    configure_file_logger(logfile, debug, 1)
  except Exception as ex:
    root_logger.exception("Logging to file is not possible. Exiting.", exc_info=ex, stack_info=True)
    sys.exit(1)

  configure_cli_logger(debug)

  flogger = get_file_logger()

  if debug:
    sys_version = sys.version.replace('\n', '')
    flogger.debug(f"CLI version: {__version__}")
    flogger.debug(f"Python version: {sys_version}")
    flogger.debug("Modules: %s", ', '.join(sorted(p.name for p in iter_modules())))

    my_system = platform.uname()
    flogger.debug(f"System: {my_system.system}")
    flogger.debug(f"Node Name: {my_system.node}")
    flogger.debug(f"Release: {my_system.release}")
    flogger.debug(f"Version: {my_system.version}")
    flogger.debug(f"Machine: {my_system.machine}")
    flogger.debug(f"Processor: {my_system.processor}")

    flogger.debug(f"Received arguments: {str(args)}")
    flogger.debug(f"Parsed arguments: {str(ns)}")

  start = perf_counter()
  success = True

  try:
    invoke_handler(ns)
  except ValueError as error:
    success = False
    logger = getLogger(__name__)
    logger.debug("ValueError occurred.", exc_info=error)
  except Exception as error:
    success = False
    logger = getLogger(__name__)
    logger.debug("Exception occurred.", exc_info=error)

  duration = perf_counter() - start
  flogger.debug(f"Total duration (seconds): {duration}")

  exit_code = 0 if success else 1
  if debug:
    # path not encapsulated in "" because it is only console out
    root_logger.info(f"See log: {logfile.absolute()}")

  sys.exit(exit_code)


def run():
  arguments = sys.argv[1:]
  parse_args(arguments)


def debug_file_exists():
  return (Path(gettempdir()) / f"{__APP_NAME}-debug").is_file()


def create_debug_file():
  if not debug_file_exists():
    (Path(gettempdir()) / f"{__APP_NAME}-debug").write_text("", "UTF-8")


if __name__ == "__main__":
  run()



import logging
from logging import getLogger
from pathlib import Path
from typing import cast

import numpy as np
import numpy.typing as npt
import torch
from tacotron import Synthesizer as TacotronSynthesizer
from tacotron import get_speaker_mapping
from tqdm import tqdm
from waveglow import Synthesizer as WaveglowSynthesizer
from waveglow import try_copy_to

from en_tts.globals import DEFAULT_CONF_DIR
from en_tts.helper import get_default_device, get_sample_count
from en_tts.resources import get_taco_model, get_wg_model


class Synthesizer():
  def __init__(
      self,
      conf_dir: Path = DEFAULT_CONF_DIR,
      device: torch.device = get_default_device()
  ) -> None:
    logger = getLogger(__name__)

    tacotron_logger = getLogger("tacotron")
    tacotron_logger.parent = logger
    tacotron_logger.setLevel(logging.WARNING)

    waveglow_logger = getLogger("waveglow")
    waveglow_logger.parent = logger
    waveglow_logger.setLevel(logging.WARNING)

    self._device = device
    self._conf_dir = conf_dir
    tacotron_ckp = get_taco_model(conf_dir, device)
    self._tacotron_ckp = tacotron_ckp
    self._tacotron = TacotronSynthesizer(
      checkpoint=tacotron_ckp,
      custom_hparams=None,
      device=device,
    )
    waveglow_ckp = get_wg_model(conf_dir, device)
    self._waveglow_ckp = waveglow_ckp
    self._waveglow = WaveglowSynthesizer(
      waveglow_ckp,
      device=device,
    )
    self._speaker = list(get_speaker_mapping(tacotron_ckp).keys())[0]
    self._paragraph_sep = "\n\n"
    self._sentence_sep = "\n"
    self._symbol_seperator = "|"

  def synthesize(self, text_ipa: str, max_decoder_steps: int = 5000, seed: int = 0, sigma: float = 1.0, denoiser_strength: float = 0.0005, silence_sentences: float = 0.2, silence_paragraphs: float = 1.0, silent: bool = False) -> npt.NDArray[np.float64]:
    resulting_wavs = []
    paragraph_sentences = [
      [
        sentence
        for sentence in paragraph.split(self._sentence_sep)
        if sentence != ""
      ]
      for paragraph in text_ipa.split(self._paragraph_sep)
    ]
    sentence_count = sum(1 for p in paragraph_sentences for s in p)

    with tqdm(desc="Synthesizing", total=sentence_count, unit=" sent", disable=silent) as pbar:
      for paragraph_nr, paragraph in enumerate(paragraph_sentences):
        for sentence_nr, sentence in enumerate(paragraph):
          # sentence_id = f"{paragraph_nr+1}-{sentence_nr+1}"
          symbols = sentence.split(self._symbol_seperator)
          # logger.debug(f"Synthesizing {sentence_id} step 1/2...")
          inf_sent_output = self._tacotron.infer(
            symbols=symbols,
            speaker=self._speaker,
            include_stats=False,
            max_decoder_steps=max_decoder_steps,
            seed=seed,
          )

          # if loglevel >= 2:
          #   logfile = work_dir / f"{sentence_id}.npy"
          #   np.save(logfile, inf_sent_output.mel_outputs_postnet)
          #   logger.debug(f"Tacotron output: {logfile.absolute()}")

          mel_var_f = torch.FloatTensor(inf_sent_output.mel_outputs_postnet)
          del inf_sent_output
          mel_var = try_copy_to(mel_var_f, self._device)
          mel_var = mel_var.unsqueeze(0)
          # logger.debug(f"Synthesizing {sentence_id} step 2/2...")
          inference_result = self._waveglow.infer(
            mel_var, sigma=sigma, denoiser_strength=denoiser_strength, seed=seed)
          # wav_inferred_denoised_normalized = normalize_wav(inference_result.wav_denoised)
          del mel_var

          # if loglevel >= 2:
          #   logfile = work_dir / f"{sentence_id}.wav"
          #   float_to_wav(wav_inferred_denoised_normalized, logfile)
          #   flogger.info(f"WaveGlow output: {logfile.absolute()}")

          resulting_wavs.append(inference_result.wav_denoised)
          is_last_sentence_in_paragraph = sentence_nr == len(paragraph) - 1
          if silence_sentences > 0 and not is_last_sentence_in_paragraph:
            pause_samples = np.zeros(
              (get_sample_count(self._waveglow.hparams.sampling_rate, silence_sentences),))
            resulting_wavs.append(pause_samples)
          pbar.update(1)

        is_last_paragraph = paragraph_nr == len(paragraph_sentences) - 1
        if silence_paragraphs > 0 and not is_last_paragraph:
          pause_samples = np.zeros(
            (get_sample_count(self._waveglow.hparams.sampling_rate, silence_paragraphs),))
          resulting_wavs.append(pause_samples)

    if len(resulting_wavs) > 0:
      resulting_wav = cast(npt.NDArray[np.float64], np.concatenate(tuple(resulting_wavs), axis=-1))
      return resulting_wav
    empty_result = cast(npt.NDArray[np.float64], np.zeros((0,), np.float64))
    return empty_result



from en_tts.transcriber import Transcriber
from en_tts_tests.helper import get_tests_dir


def test_component():
  conf_dir = get_tests_dir()

  t = Transcriber(conf_dir)

  text = "This is a test abbb? And I'm there 31."
  text_ipa = t.transcribe_to_ipa(text, False, False)

  assert text_ipa == 'ð|ˈɪ|s|SIL0|ˈɪ|z|SIL0|ə|SIL0|tː|ˈɛ|s|t|SIL0|ˈæ|b|?|SIL2\nə|n˘|d|SIL0|ˈaɪ˘|m|SIL0|ð|ˈɛr˘|SIL0|θ|ˈʌr|d˘|ˌi|-|wː|ˈʌː|nː|.|SIL2'



import argparse
import codecs
from argparse import ArgumentTypeError
from functools import partial
from pathlib import Path
from typing import Callable, Optional, Sequence, TypeVar

import torch
from ordered_set import OrderedSet

T = TypeVar("T")


class ConvertToOrderedSetAction(argparse._StoreAction):
  def __call__(self, parser: argparse.ArgumentParser, namespace: argparse.Namespace, values: Optional[Sequence], option_string: Optional[str] = None):
    # Note: normal set is not possible because set is not a Sequence
    val: Optional[OrderedSet] = None
    if values is not None:
      val = OrderedSet(values)
    super().__call__(parser, namespace, val, option_string)


def parse_device(value: str) -> torch.device:
  try:
    device = torch.device(value)
  except Exception as ex:
    raise ArgumentTypeError("Device was not found!") from ex
  return device


def get_torch_devices():
  yield "cpu"
  cuda_count = torch.cuda.device_count()
  if cuda_count == 1:
    yield "cuda"
  else:
    for i in range(cuda_count):
      yield f"cuda:{i}"


def parse_codec(value: str) -> str:
  value = parse_required(value)
  try:
    codecs.lookup(value)
  except LookupError as error:
    raise ArgumentTypeError("Codec was not found!") from error
  return value


def parse_path(value: str) -> Path:
  value = parse_required(value)
  try:
    path = Path(value)
  except ValueError as error:
    raise ArgumentTypeError("Value needs to be a path!") from error
  return path


def parse_optional_value(value: str, method: Callable[[str], T]) -> Optional[T]:
  if value is None:
    return None
  return method(value)


def parse_float_between_zero_and_one(value: str) -> float:
  result = parse_float(value)
  if not 0 <= result <= 1:
    raise ArgumentTypeError("Value needs to be in interval [0, 1]!")
  return result


def parse_character(value: str) -> str:
  if len(value) != 1:
    raise ArgumentTypeError("Value needs to be one character!")
  return value


def get_optional(method: Callable[[str], T]) -> Callable[[str], Optional[T]]:
  result = partial(
    parse_optional_value,
    method=method,
  )
  return result


def parse_existing_file(value: str) -> Path:
  path = parse_path(value)
  if not path.is_file():
    raise ArgumentTypeError("File was not found!")
  return path


def parse_existing_directory(value: str) -> Path:
  path = parse_path(value)
  if not path.is_dir():
    raise ArgumentTypeError("Directory was not found!")
  return path


def parse_required(value: Optional[str]) -> str:
  if value is None:
    raise ArgumentTypeError("Value must not be None!")
  return value


def parse_non_empty(value: Optional[str]) -> str:
  value = parse_required(value)
  if value == "":
    raise ArgumentTypeError("Value must not be empty!")
  return value


def parse_non_empty_or_whitespace(value: str) -> str:
  value = parse_required(value)
  if value.strip() == "":
    raise ArgumentTypeError("Value must not be empty or whitespace!")
  return value


def parse_float(value: str) -> float:
  value_str = parse_required(value)
  try:
    result = float(value_str)
  except ValueError as error:
    raise ArgumentTypeError("Value needs to be a decimal number!") from error
  return result


def parse_positive_float(value: str) -> float:
  result = parse_float(value)
  if not result > 0:
    raise ArgumentTypeError("Value needs to be greater than zero!")
  return result


def parse_non_negative_float(value: str) -> float:
  result = parse_float(value)
  if not result >= 0:
    raise ArgumentTypeError("Value needs to be greater than or equal to zero!")
  return result


def parse_integer(value: str) -> int:
  value_str = parse_required(value)
  if not value_str.isdigit():
    raise ArgumentTypeError("Value needs to be an integer!")
  result = int(value_str)
  return result


def parse_positive_integer(value: str) -> int:
  result = parse_integer(value)
  if not result > 0:
    raise ArgumentTypeError("Value needs to be greater than zero!")
  return result


def parse_integer_greater_one(value: str) -> int:
  result = parse_integer(value)
  if not result > 1:
    raise ArgumentTypeError("Value needs to be greater than one!")
  return result


def parse_float_greater_one(value: str) -> float:
  result = parse_float(value)
  if not result > 1:
    raise ArgumentTypeError("Value needs to be greater than one!")
  return result


def parse_non_negative_integer(value: str) -> int:
  result = parse_integer(value)
  if not result >= 0:
    raise ArgumentTypeError("Value needs to be greater than or equal to zero!")
  return result



from logging import getLogger
from pathlib import Path
from typing import cast

import nltk
import torch
import wget
from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      PronunciationDict, load_dict)
from tacotron import CheckpointDict
from tacotron_cli import load_checkpoint
from waveglow import CheckpointWaveglow, convert_glow_files
from waveglow_cli import download_pretrained_model

from en_tts.io import load_obj, save_obj

LJS_DUR_DICT = "https://zenodo.org/record/7499098/files/pronunciations.dict"
CMU_IPA_DICT = "https://zenodo.org/record/7500805/files/pronunciations.dict"
TACO_CKP = "https://zenodo.org/records/10107104/files/101000.pt"


def get_ljs_dict(conf_dir: Path) -> PronunciationDict:
  logger = getLogger(__name__)
  conf_dir.mkdir(parents=True, exist_ok=True)
  ljs_dict_path = conf_dir / "ljs.dict"
  ljs_dict_pkl_path = conf_dir / "ljs.dict.pkl"

  ljs_dict: PronunciationDict
  if not ljs_dict_path.is_file():
    logger.info("Downloading LJS dictionary ...")
    wget.download(LJS_DUR_DICT, str(ljs_dict_path.absolute()))
    logger.info("Loading LJS dictionary ...")
    # 78k lines
    ljs_dict = load_dict(ljs_dict_path, "UTF-8", DeserializationOptions(
      False, False, False, True), MultiprocessingOptions(1, None, 100_000))
    save_obj(ljs_dict, ljs_dict_pkl_path)
  else:
    logger.info("Loading LJS dictionary ...")
    ljs_dict = cast(PronunciationDict, load_obj(ljs_dict_pkl_path))
  return ljs_dict


def get_cmu_dict(conf_dir: Path) -> PronunciationDict:
  logger = getLogger(__name__)
  conf_dir.mkdir(parents=True, exist_ok=True)
  cmu_dict_path = conf_dir / "cmu.dict"
  cmu_dict_pkl_path = conf_dir / "cmu.dict.pkl"

  cmu_dict: PronunciationDict
  if not cmu_dict_path.is_file():
    logger.info("Downloading CMU dictionary ...")
    wget.download(CMU_IPA_DICT, str(cmu_dict_path.absolute()))

    logger.info("Loading CMU dictionary ...")
    cmu_dict = load_dict(cmu_dict_path, "UTF-8", DeserializationOptions(
      False, False, False, False), MultiprocessingOptions(1, None, 100_000))
    save_obj(cmu_dict, cmu_dict_pkl_path)
  else:
    cmu_dict = cast(PronunciationDict, load_obj(cmu_dict_pkl_path))
  return cmu_dict


def get_wg_model(conf_dir: Path, device: torch.device) -> CheckpointWaveglow:
  logger = getLogger(__name__)
  conf_dir.mkdir(parents=True, exist_ok=True)
  wg_path = conf_dir / "waveglow.pt"
  wg_orig_path = conf_dir / "waveglow_orig.pt"

  if not wg_path.is_file():
    logger.info("Downloading WaveGlow checkpoint ...")
    download_pretrained_model(wg_orig_path, version=3)
    wg_checkpoint = convert_glow_files(wg_orig_path, wg_path, device, keep_orig=False)
    # wget.download(WG_CKP, str(wg_path.absolute()))
  else:
    logger.info("Loading WaveGlow checkpoint ...")  # from: {wg_path.absolute()} ...")
    wg_checkpoint = CheckpointWaveglow.load(wg_path, device)
  return wg_checkpoint


def get_taco_model(conf_dir: Path, device: torch.device) -> CheckpointDict:
  logger = getLogger(__name__)
  conf_dir.mkdir(parents=True, exist_ok=True)
  taco_path = conf_dir / "tacotron.pt"

  if not taco_path.is_file():
    logger.info("Downloading Tacotron checkpoint ...")
    wget.download(TACO_CKP, str(taco_path.absolute()))

  logger.info("Loading Tacotron checkpoint ...")  # from: {taco_path.absolute()} ...")
  checkpoint = load_checkpoint(taco_path, device)
  return checkpoint


def download_nltk_data():
  logger = getLogger(__name__)
  try:
    nltk.data.find('taggers/averaged_perceptron_tagger.zip')
  except LookupError:
    logger.info("Downloading 'averaged_perceptron_tagger' from nltk ...")
    nltk.download('averaged_perceptron_tagger')
  try:
    nltk.data.find('corpora/cmudict.zip')
  except LookupError:
    logger.info("Downloading 'cmudict' from nltk ...")
    nltk.download('cmudict')
